@@ -12,6 +12,21 @@ import StringIO
 import termios
 import sys
 import re
+import functools
+
+
+def memoize(obj):
+# from https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
+    cache = obj.cache = {}
+
+    @functools.wraps(obj)
+    def memoizer(*args, **kwargs):
+        key = str(args) + str(kwargs)
+        if key not in cache:
+            cache[key] = obj(*args, **kwargs)
+        return cache[key]
+    return memoizer
+
 
 class Shell(shellpic.Formatter):
     """
@@ -206,19 +221,20 @@ class Shell(shellpic.Formatter):
         file_str.write(chr(27) + u"[0m")
         return file_str.getvalue()
 
-class Shell8bit(Shell):
+class Shell8Bit(Shell):
     """
     A formatter designed for terminals capable of showing 256-colors
     (e.g. xterm and gnome-terminal).
 
     """
     def __init__(self):
-        super(Shell8bit, self).__init__()
+        super(Shell8Bit, self).__init__()
 
-    @classmethod
-    def colorcode(cls, bgcolor, fgcolor):
-        return u"{}[48;5;{};38;5;{}m{}▄ ".format(chr(27), cls.color_value_8bit(*bgcolor),
-                                                 cls.color_value_8bit(*fgcolor), chr(8))
+    @staticmethod
+    @memoize
+    def colorcode(bgcolor, fgcolor):
+        return u"{}[48;5;{};38;5;{}m{}▄ ".format(chr(27), Shell8Bit.color_value_8bit(*bgcolor),
+                                                 Shell8Bit.color_value_8bit(*fgcolor), chr(8))
 
     @staticmethod
     def color_value_8bit(r, g, b, a=255):
@@ -244,7 +260,72 @@ class Shell24Bit(Shell):
     def __init__(self):
         super(Shell24Bit, self).__init__()
 
-    @classmethod
-    def colorcode(cls, bgcolor, fgcolor):
+    @staticmethod
+    @memoize
+    def colorcode(bgcolor, fgcolor):
         return u"{}[48;2;{};{};{};38;2;{};{};{}m{}▄ ".format(chr(27), bgcolor[0], bgcolor[1], bgcolor[2],
                                                             fgcolor[0], fgcolor[1], fgcolor[2], chr(8))
+
+class Shell4Bit(Shell):
+    """
+    A formatter for 16-color terminals.
+    """
+
+    palette = (
+        (0, 0, 0),       #  0 black
+        (205, 0, 0),     #  1 red3
+        (0, 205, 0),     #  2 green3
+        (205, 205, 0),   #  3 yellow3
+        (0, 0, 205),     #  4 blue3
+        (205, 0, 205),   #  5 magenta3
+        (0, 205, 205),   #  6 cyan
+        (229, 229, 229), #  7 gray90
+        (77, 77, 77),    #  8 gray30
+        (255, 0, 0),     #  9 red
+        (0, 255, 0),     # 10 green
+        (255, 255, 0),   # 11 yellow
+        (0, 0, 255),     # 12 blue
+        (255, 0, 255),   # 13 magenta
+        (0, 255, 255),   # 14 cyan
+        (255, 255, 255), # 15 white
+        )
+
+    weights = (
+        0.05,       #  0 black
+        0.25,     #  1 red3
+        1.2,     #  2 green3
+        0.5,   #  3 yellow3
+        1,     #  4 blue3
+        0.5,   #  5 magenta3
+        1,   #  6 cyan
+        0.12, #  7 gray90
+        0.12,    #  8 gray30
+        0.5,     #  9 red
+        1,     # 10 green
+        0.75,   # 11 yellow
+        1.25,     # 12 blue
+        0.75,   # 13 magenta
+        1.5,   # 14 cyan
+        0.5, # 15 white
+        )
+
+    def __init__(self):
+        super(Shell4Bit, self).__init__()
+
+    @classmethod
+    def color_value_4bit(cls, r, g, b, a=255):
+        def distance(a, b):
+            return sum([pow(x - y, 2) for x, y in zip(a, b)])
+        distances = [[distance(p, [r, g, b]), i] for i, p in enumerate(cls.palette)]
+        for d in distances:
+            d[0] /= cls.weights[d[1]]
+        distances.sort(key=lambda x: x[0])
+        code = distances[0][1]
+        code = 30 + code if code < 8 else 82 + code
+        return code
+
+    @staticmethod
+    @memoize
+    def colorcode(bgcolor, fgcolor):
+        return u"{}[{};{}m▄ ".format(chr(27), Shell4Bit.color_value_4bit(*bgcolor) + 10, 
+                                     Shell4Bit.color_value_4bit(*fgcolor))
